@@ -1,5 +1,6 @@
+# vim: set fileencoding=utf-8:
 # GNU Solfege - free ear training software
-# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  Tom Cato Amundsen
+# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011  Tom Cato Amundsen
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,14 +18,15 @@
 from __future__ import absolute_import
 
 import errno
-import gobject
 import locale
 import logging
 import os
 import shutil
+import sqlite3
 import subprocess
 import sys
 import tempfile
+from urlparse import urlparse
 import webbrowser
 
 from solfege import mpd
@@ -33,26 +35,51 @@ from solfege import soundcard
 from solfege import abstract
 from solfege import cfg
 from solfege import dataparser
+from solfege import filesystem
 from solfege import gu
 from solfege import i18n
 from solfege import lessonfile
 from solfege import osutils
+from solfege import parsetree
 from solfege import reportlib
 from solfege import utils
 
 import solfege
 
+try:
+    from pyalsa import alsaseq
+except ImportError:
+    alsaseq = None
+
+solfege_copyright = u"Copyright © 1999-2008 Tom Cato Amundsen <tca@gnu.org>, and others."
+
+warranty = """
+This program is free software: you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation, either version 3 of the License, or
+(at your option) any later version.
+
+This program is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this program.  If not, see <http://www.gnu.org/licenses/>.
+"""
+
+
 def check_rcfile():
     """See default.config for rcfileversion values, meanings and
     a description of how to add config variables.
     """
-    rcfileversion = 19
+    rcfileversion = 20
     if cfg.get_int("app/rcfileversion") > rcfileversion:
         cfg.drop_user_config()
         return
     if cfg.get_int("app/rcfileversion") <= 1:
         if not "example-files" in cfg.get_string('config/lessoncollections'):
-            cfg.set_string('config/lessoncollections', 
+            cfg.set_string('config/lessoncollections',
                 "%s example-files" % cfg.get_string('config/lessoncollections'))
     if cfg.get_int("app/rcfileversion") <= 5:
         # This is more complicated that necessary to fix an old
@@ -118,8 +145,8 @@ def check_rcfile():
                 " ".join(v[1:]))
     if cfg.get_int("app/rcfileversion") < 17:
         v = cfg.get_string("app/frontpage").split("/")
-        if v[0] == u"exercises":
-            cfg.set_string("app/frontpage", 
+        if v[0] == u"exercises" and v[1] != u"standard":
+            cfg.set_string("app/frontpage",
                            u"/".join([v[0], u"standard"] + v[1:]))
     if cfg.get_int("app/rcfileversion") < 18:
         cfg.del_key("gui/web_browser_as_help_browser")
@@ -127,6 +154,9 @@ def check_rcfile():
         for ex in ('singinterval', 'melodicinterval'):
             if cfg.get_int("%s/maximum_number_of_intervals" % ex) == 10:
                 cfg.set_int("%s/maximum_number_of_intervals" % ex, 12)
+    if cfg.get_int("app/rcfileversion") < 20:
+        cfg.del_key("gui/reserved_vspace")
+
     cfg.set_int("app/rcfileversion", rcfileversion)
     try:
         a = mpd.notename_to_int(cfg.get_string("user/lowest_pitch"))
@@ -168,6 +198,29 @@ class SolfegeApp(cfg.ConfigUtils):
         if self.m_options.no_sound \
            or cfg.get_string("sound/type") == "fake-synth":
             soundcard.initialise_using_fake_synth(self.m_options.verbose_sound_init)
+        elif cfg.get_string("sound/type") == "alsa-sequencer":
+            if alsaseq:
+                try:
+                    clientid, portid = self.get_list("sound/alsa-client-port")
+                except ValueError:
+                    clientid, portid = (None, None)
+                try:
+                    soundcard.initialise_alsa_sequencer((clientid, portid),
+                            self.m_options.verbose_sound_init)
+                except alsaseq.SequencerError, e:
+                    logging.debug("initialise_alsa_sequencer failed. Using fake synth.")
+                    self.display_sound_init_error_message(e)
+                    soundcard.initialise_using_fake_synth(True)
+                    return
+            else:
+                if solfege.splash_win:
+                    solfege.splash_win.hide()
+                gu.dialog_ok(_("The pyalsa Python module is missing"),
+                    solfege.win,
+                    _("Solfege was configured to use the Python modules from www.alsa-project.org, but the modules were not found. You must reconfigure sound in the preferences window (Ctrl-F12) or restart Solfege in a way that it finds the modules."))
+                soundcard.initialise_using_fake_synth(True)
+                if solfege.splash_win:
+                    solfege.splash_win.show()
         elif cfg.get_string("sound/type") == "winsynth":
             try:
                 soundcard.initialise_winsynth(cfg.get_int("sound/synth_number"),
@@ -181,7 +234,7 @@ class SolfegeApp(cfg.ConfigUtils):
                 # We can get here if winmidi.output_devices() in winsynth
                 # __init__ returns no devices. Don't know when, but it could
                 # happen.
-                solfege.win.display_exception_message(e)
+                gu.display_exception_message(e)
                 cfg.set_string("sound/type", "fake-synth")
                 soundcard.initialise_using_fake_synth(True)
                 return
@@ -190,7 +243,6 @@ class SolfegeApp(cfg.ConfigUtils):
                 cfg.set_int("sound/synth_number", soundcard.synth.m_devnum)
         elif cfg.get_string("sound/type") == "external-midiplayer":
             soundcard.initialise_external_midiplayer(
-                    cfg.get_string("sound/midi_player"),
                     verbose_init=self.m_options.verbose_sound_init)
             soundcard.synth.error_report_cb = solfege.win.display_error_message
         elif cfg.get_string("sound/type") == '':
@@ -228,13 +280,13 @@ class SolfegeApp(cfg.ConfigUtils):
             """%s""" % str(e).decode(locale.getpreferredencoding(), 'replace'))
         elif isinstance(e, ImportError):
             solfege.win.display_error_message2(str(e), _("You should configure sound from the preferences window, and try to use an external midi player. Or try to recompile the program and check for error messages to see why the module is not built."))
-        elif e.errno == errno.EACCES:
+        elif getattr(e, 'errno', None) == errno.EACCES:
             solfege.win.display_error_message(
                 "The sound init failed: %s\n"
                 "The errno EACCES indicates that you don't have write "
                 "permission to the device."
                 % str(e).decode(locale.getpreferredencoding(), 'replace'))
-        elif e.errno == errno.EBUSY:
+        elif getattr(e, 'errno', None) == errno.EBUSY:
             solfege.win.display_error_message(
                 "The sound init failed: %s\n"
                 "It seems like some other program is using the device. You "
@@ -265,7 +317,7 @@ class SolfegeApp(cfg.ConfigUtils):
                 self.handle_href('%s.html' % solfege.win.m_viewer)
     def show_exercise_theory(self):
         if self.m_teachers[self.m_running_exercise].m_P.header.theory:
-            solfege.win.display_docfile("%s.html" % self.m_teachers[self.m_running_exercise].m_P.header.theory, "")
+            solfege.win.display_docfile("%s.html" % self.m_teachers[self.m_running_exercise].m_P.header.theory)
     def _practise_lessonfile(self, filename, urlobj=None):
         """
         return the module name.
@@ -273,6 +325,15 @@ class SolfegeApp(cfg.ConfigUtils):
         module = lessonfile.infocache.get(filename, 'module')
         if self.m_running_exercise:
             solfege.win.box_dict[self.m_running_exercise].on_end_practise()
+        if not lessonfile.is_uri(filename):
+            # Since the file is in ~/.solfege/exercises we must check
+            # if the user have written his own exercise module
+            if os.path.exists(os.path.normpath(os.path.join(
+                            os.path.dirname(filename),
+                            "..", "modules", u"%s.py" % module))):
+                module = u"user:%s/%s" % (
+                    os.path.dirname(filename).split(os.sep)[-2],
+                    module)
         if module not in self.m_teachers:
             self.create_teacher(module)
         if module not in solfege.win.box_dict:
@@ -295,9 +356,11 @@ class SolfegeApp(cfg.ConfigUtils):
         try:
             module = self._practise_lessonfile(filename)
         except (lessonfile.LessonfileParseException,
-                dataparser.DataparserException), e:
+                dataparser.DataparserException,
+                parsetree.ParseTreeException,
+                IOError), e:
             cleanup()
-            solfege.win.display_exception_message(e, lessonfile=filename)
+            gu.display_exception_message(e, lessonfile=filename)
             return
         if 'm_discards' in dir(self.m_teachers[module].m_P):
             for msg in self.m_teachers[module].m_P.m_discards:
@@ -313,23 +376,49 @@ class SolfegeApp(cfg.ConfigUtils):
     def test_lessonfile(self, filename):
         self.m_test_mode = True
         module = self.practise_lessonfile(filename)
-        solfege.win.box_dict[module].on_start_practise()
         solfege.win.enter_test_mode()
-    def handle_href(self, href, display_docfile_set_adj=1):
-        urlobj = utils.Url(href)
-        if not urlobj.protocol:
-            solfege.win.display_docfile(urlobj.filename, urlobj.anchor)
-        else:
+    def handle_href(self, href):
+        u = urlparse(href)
+        if u.scheme:
             try:
                 webbrowser.open_new(href)
             except Exception, e:
                 solfege.win.display_error_message2(_("Error opening web browser"), str(e))
+        else:
+            solfege.win.display_docfile(u.path)
     def create_teacher(self, modulename):
         """
         Create the teacher in 'modulename' and add it to self.m_teachers.
         """
-        m = __import__("solfege.exercises.%s" % modulename, fromlist=("solfege.exercises.%s" % modulename,), level=0)
+        m = self.import_module(modulename)
         self.m_teachers[modulename] = m.Teacher(modulename)
+    def import_module(self, modulename):
+        """
+        If prefixed with "solfege:"
+
+          user:collection/modulename
+
+        collection is the directory name in
+        ~/.solfege/exercises/collection/modulename
+        and "user:" is just a prefix to show that the module name
+        is in the users directory.
+        
+        If a plain string with not prefix, it is one of the standard modules
+        included with Solfege.
+
+        Return the imported module
+        """
+        if modulename.startswith("user:"):
+            collection = modulename[len("user:"):].split(os.sep)[0]
+            module_dir = os.path.join(filesystem.user_data(),
+                                      "exercises", collection, "modules")
+            sys.path.insert(0, module_dir)
+            m = __import__(modulename.split("/")[1])
+            reload(m)
+            del sys.path[0]
+        else:
+            m = __import__("solfege.exercises.%s" % modulename, fromlist=("solfege.exercises.%s" % modulename,), level=0)
+        return m
     def reset_exercise(self, w=None):
         """
         Call on_end_practise, and then on_start_practise in
@@ -339,23 +428,31 @@ class SolfegeApp(cfg.ConfigUtils):
             solfege.win.get_view().on_end_practise()
             solfege.win.get_view().on_start_practise()
     def quit_program(self):
-        try:
-            cfg.sync()
-        except IOError, e:
-            solfege.win.display_exception_message(e)
         if isinstance(solfege.win.get_view(), abstract.Gui):
             g = solfege.win.get_view()
             # Check that the lesson file has a header, because if the
             # user tries to quit the program after parsing a bad lesson
             # file, we cannot call end_practise() without risking more
             # exceptions.
-            if g.m_t.m_P and getattr(g.m_t.m_P, 'header', None):
+            if g.m_t.m_P and hasattr(g.m_t.m_P, 'header'):
                 g.on_end_practise()
-        solfege.db.conn.close()
+        try:
+            cfg.sync()
+        except IOError, e:
+            gu.display_exception_message(e)
+        try:
+            solfege.db.conn.commit()
+        except sqlite3.ProgrammingError, e:
+            gu.display_exception_message(e)
+        try:
+            solfege.db.conn.close()
+        except sqlite3.ProgrammingError, e:
+            gu.display_exception_message(e)
         if soundcard.synth:
             soundcard.synth.close()
         shutil.rmtree(lessonfile.MusicBaseClass.temp_dir, True)
-    def export_training_set(self, export_data, export_dir, output_format):
+    def export_training_set(self, export_data, export_dir, output_format,
+                            name_track_by_question):
         """
         This function requires a program that can create WAV files
         from MIDI files and MP3 files from WAV.
@@ -385,6 +482,7 @@ class SolfegeApp(cfg.ConfigUtils):
             p = self.m_teachers[module].lessonfileclass()
             p.parse_file(lessonfile.uri_expand(filename))
             for c in range(lesson_info['count']):
+                trackname = "track-%i"
                 if module == 'idbyname':
                     p.select_random_question()
                     if p.header.lesson_heading:
@@ -394,8 +492,10 @@ class SolfegeApp(cfg.ConfigUtils):
                     table.append_row("%i" % track_idx,
                                      p.get_question().name,
                                      s)
+                    if name_track_by_question:
+                        trackname = "%s-%%i" % p.get_name()
                     soundcard.start_export(os.path.join(
-                            export_dir, "track-%i.mid" % track_idx))
+                            export_dir, "%s.mid" % trackname % track_idx))
                     for n in range(lesson_info.get('repeat', 1)):
                         p.play_question()
                         if n != lesson_info.get('repeat', 1) - 1:
@@ -410,10 +510,14 @@ class SolfegeApp(cfg.ConfigUtils):
                     t.q_status = t.QSTATUS_SOLVED
                     try:
                         table.append_row("%i" % track_idx, "%s" % utils.int_to_intervalname(t.m_interval))
+                        if name_track_by_question:
+                            trackname = "%%i-%s.mid" % utils.int_to_intervalname(t.m_interval)
                     except AttributeError:
                         table.append_row("%i" % track_idx, "%s" % (" + ".join([utils.int_to_intervalname(q, False, True) for q in t.m_question])))
+                        if name_track_by_question:
+                            trackname = "%%i-%s.mid" % ("+".join([utils.int_to_intervalname(q, False, True) for q in t.m_question]))
                     soundcard.start_export(os.path.join(
-                            export_dir, "track-%i.mid" % track_idx))
+                            export_dir, "%s.mid" % trackname % track_idx))
                     for n in range(lesson_info.get('repeat', 1)):
                         t.play_question()
                         if n != lesson_info.get('repeat', 1) - 1:
@@ -422,7 +526,7 @@ class SolfegeApp(cfg.ConfigUtils):
                                     (self.get_int('config/default_bpm'), 4))
                     soundcard.end_export()
                 else:
-                    logging.warning("export_training_set:ignoring exercise with module='%s'" % module)
+                    logging.warning("export_training_set:ignoring exercise with module='%s'", module)
 #####
                 def do_convert(from_format, to_format):
                     """
@@ -440,9 +544,9 @@ class SolfegeApp(cfg.ConfigUtils):
                     try:
                         inout = {
                             'in': os.path.join(export_dir,
-                                    "track-%i.%s" % (track_idx, from_ext)),
+                                    "%s.%s" % (trackname % track_idx, from_ext)),
                             'out': os.path.join(export_dir,
-                                    "track-%i.%s" % (track_idx, to_ext))}
+                                    "%s.%s" % (trackname % track_idx, to_ext))}
                         opts = cfg.get_string(app_cfg_name + '_options').split(" ")
                         opts = [x % inout for x in opts]
                         # For some reasong setting the executable arg does
@@ -454,8 +558,8 @@ class SolfegeApp(cfg.ConfigUtils):
                             raise osutils.BinaryForMediaConvertorException(app_cfg_name,
                                 cfg.get_string(app_cfg_name), e)
 
-                        if os.path.exists(os.path.join(export_dir, "track-%i.%s" % (track_idx, to_ext))):
-                            os.remove(os.path.join(export_dir, "track-%i.%s" % (track_idx, from_ext)))
+                        if os.path.exists(os.path.join(export_dir, "%s.%s" % (trackname % track_idx, to_ext))):
+                            os.remove(os.path.join(export_dir, "%s.%s" % (trackname % track_idx, from_ext)))
                         else:
                             # This means that the program failed to generate
                             # the WAV file. We set output_format to 'midi'
